@@ -1,7 +1,7 @@
 package Text::MediawikiFormat;
 
-use warnings;
 use strict;
+use warnings::register;
 
 =head1 NAME
 
@@ -9,15 +9,16 @@ Text::MediawikiFormat - Translate Wiki markup into other text formats
 
 =head1 VERSION
 
-Version 0.06
+Version 1.0
 
 =cut
 
-our $VERSION = '0.06';
+use vars qw($VERSION);
+use version; $VERSION = qv('1.0');
 
 =head1 SYNOPSIS
 
-	use Text::MediawikiFormat;
+	use Text::MediawikiFormat 'wikiformat';
 	my $html = wikiformat ($raw);
 	my $text = wikiformat ($raw, {}, {implicit_links => 1});
 
@@ -32,7 +33,9 @@ default, but could produce valid POD, DocBook, XML, or any other format
 imaginable.
 
 The most important function is C<Text::MediawikiFormat::format()>.  It is
-exported as C<wikiformat()> by default, but that function name may be changed.
+not exported by default, but will be exported as C<wikiformat()> if any
+options at all are passed to the exporter, unless the name is overridden
+explicitly.  See L<"EXPORT"> for more information.
 
 It should be noted that this module is written as a drop in replacement for
 L<Text::WikiMarkup> that expands on that modules functionality and provides
@@ -48,9 +51,11 @@ use CGI qw(:standard);
 use Scalar::Util qw(blessed);
 use Text::MediawikiFormat::Blocks;
 use URI;
-use URI::Escape;
+use URI::Escape qw(uri_escape uri_escape_utf8);
+
+use vars qw($missing_html_packages %tags %opts %merge_matrix
+	    $uric $uricCheat $uriCruft);
 
-our $missing_html_packages;
 BEGIN
 {
     # Try to load optional HTML packages, recording any errors.
@@ -65,7 +70,7 @@ BEGIN
 ###
 ### Defaults
 ###
-our %tags =
+%tags =
 (
     indent		=> qr/^(?:[:*#;]*)(?=[:*#;])/,
     link		=> \&_make_html_link,
@@ -134,13 +139,14 @@ our %tags =
     _toc		=> [],
 );
 
-our %opts =
+%opts =
 (
     extended       => 1,
     implicit_links => 0,
     absolute_links => 1,
     prefix         => '',
     process_html   => 1,
+    charset	   => 'utf-8',
 );
 
 # Make sure import's argument hash contains an `as' entry.  `as' defaults to
@@ -172,7 +178,7 @@ sub _extract_opts
 
 # Shamelessly ripped from Hash::Merge, which doesn't work in a threaded
 # environment with two threads trying to use different merge matrices.
-our %merge_matrix =
+%merge_matrix =
 (
     SCALAR =>
     {
@@ -440,15 +446,24 @@ sub _make_schema_regex
     return qr/(?:$re)/;
 }
 
-our $uric = $URI::uric;
-our $uricCheat = $uric;
+$uric = $URI::uric;
+$uricCheat = $uric;
 
 # We need to avoid picking up 'HTTP::Request::Common' so we have a
 # subset of uric without a colon.
 $uricCheat =~ tr/://d;
 
 # Identifying characters often accidentally picked up trailing a URI.
-our $uriCruft = q/]),.!'";}/;
+$uriCruft = q/]),.!'";}/;
+
+# escape a URI based on our charset.
+sub _escape_uri
+{
+    my ($opts, $uri) = @_;
+    confess "charset not initialized" unless $opts->{charset};
+    return uri_escape_utf8 $uri if $opts->{charset} =~ /^utf-?8$/i;
+    return uri_escape $uri;
+}
 
 # Turn [[Wiki Link|Title]], [URI Title], scheme:url, or StudlyCaps into links.
 sub _make_html_link
@@ -460,8 +475,8 @@ sub _make_html_link
     if ($tag =~ /^\[\[([^|#]*)(?:(#)([^|]*))?(?:(\|)(.*))?\]\]$/)
     {
 	# Wiki link
-	$href = $opts->{prefix} . uri_escape $1 if $1;
-	$href .= $2 . uri_escape $3 if $2;
+	$href = $opts->{prefix} . _escape_uri $opts, $1 if $1;
+	$href .= $2 . _escape_uri $opts, $3 if $2;
 
 	if ($4)
 	{
@@ -505,18 +520,17 @@ sub _make_html_link
 	$tags->{_schema_regex} ||= _make_schema_regex @{$tags->{schemas}};
 	my $s = $tags->{_schema_regex};
 
-	if ($tag =~ /^(?:&lt;)?$s:[$uricCheat][$uric]*(?:&gt;)?$/)
+	if ($tag =~ /^$s:[$uricCheat][$uric]*$/)
 	{
 	    # absolute link
 	    $href = $&;
 	    $trailing = $& if $href =~ s/[$uriCruft]$//;
-	    $href =~ s/^(?:&lt;)?(.+?)(?:&gt;)?$/$1/;
 	    $title = $href;
 	}
 	else
 	{
 	    # StudlyCaps
-	    $href = $opts->{prefix} . uri_escape $tag;
+	    $href = $opts->{prefix} . _escape_uri $opts, $tag;
 	    $title = $tag;
 	}
     }
@@ -556,7 +570,7 @@ sub _store_toc_line
 sub _make_header
 {
     my $level = length $_[2];
-    my $n = uri_escape $_[3];
+    my $n = _escape_uri $_[-1], $_[3];
 
     _store_toc_line ($_[-2]->{_toc}, $level, $_[3], $n);
 
@@ -624,7 +638,7 @@ sub _append_processed_line
     if (@$lines && $lines->[-1]->[1] !~ /\n$/
 	&& # State not changing from or to 'nowiki'
 	   !($state ne $lines->[-1]->[0]
-	     && ($state eq 'nowiki' || $lines->[-1]->[0] eq 'nowiki')))
+	     && grep /^nowiki$/, $state, $lines->[-1]->[0]))
     {
 	$lines->[-1]->[1] .= shift @newlines;
 	$lines->[-1]->[0] = $state if $state eq 'html';
@@ -644,6 +658,11 @@ sub _html_tag
     my ($parser, $type, $tagname, $orig, $attr) = @_;
     my $tags = $parser->{tags};
 
+    # $tagname may have been generated by an empty tag.  If so, HTML::Parser
+    # will sometimes include the trailing / in the tag name.
+    my $isEmptyTag = $orig =~ m#/>$#;
+    $tagname =~ s#/$## if $isEmptyTag;
+
     unless (grep /^\Q$tagname\E$/, @{$tags->{allowed_tags}})
     {
 	_append_processed_line $parser, CGI::escapeHTML $orig;
@@ -652,10 +671,13 @@ sub _html_tag
     # Any $tagname must now be in the allowed list, including <nowiki>.
 
     my $tagstack = $parser->{tag_stack};
+    my $stacktop = @$tagstack ? $tagstack->[-1] : '';
 
     # First, process end tags, since they can change our state.
-    if ($type eq 'E' && @$tagstack && $tagstack->[-1] eq $tagname)
+    if ($type eq 'E' && $stacktop eq $tagname)
     {
+	# The closing tag is at the top of the stack, like it should be.
+	# Pop it and append the close tag to the output.
 	pop @$tagstack;
 	my $newtag;
 
@@ -670,40 +692,31 @@ sub _html_tag
 	}
 
 	# Can't close a state into <pre> or <nowiki>
-	my $newstate;
-	$newstate = 'html'
-	    unless $HTML::Tagset::isPhraseMarkup{$tagname};
+	_append_processed_line $parser, $newtag, 'html';
+	return;
+    }
 
-	_append_processed_line $parser, $newtag, $newstate;
+    if (@$tagstack && grep /^\Q$stacktop\E$/, qw{nowiki pre})
+    {
+	# Ignore all markup within <pre> or <nowiki> tags.
+	_append_processed_line $parser, CGI::escapeHTML ($orig), 'nowiki';
 	return;
     }
 
     if ($type eq 'E' && $HTML::Tagset::isPhraseMarkup{$tagname})
+	# If we ask for artificial end element events for self-closed elements,
+	# then we need to check $HTML::Tagset::emptyElement($tagname) here too.
     {
+	# We didn't record phrase markup on the stack, so it's okay to just
+	# let it close.
 	_append_processed_line $parser, "</$tagname>";
 	return;
     }
 
-    if ($type eq 'E'
-	|| @$tagstack
-	   && ($tagstack->[-1] eq 'pre' || $tagstack->[-1] eq 'nowiki'))
+    if ($type eq 'E')
     {
-	# We just tried to close a non-phrase that isn't on the
-	# stack and/or the current state is <pre> or <nowiki>.
-	my $newstate;
-	if (@$tagstack)
-	{
-	    if ($tagstack->[-1] eq 'pre')
-	    {
-		$newstate = 'html';
-	    }
-	    elsif ($tagstack->[-1] eq 'nowiki')
-	    {
-		$newstate = 'nowiki';
-	    }
-	}
-
-	_append_processed_line $parser, CGI::escapeHTML ($orig), $newstate;
+	# We got a non-phrase end tag that wasn't on the stack.  Escape it.
+	_append_processed_line $parser, CGI::escapeHTML ($orig);
 	return;
     }
 
@@ -715,7 +728,8 @@ sub _html_tag
     # The browser doesn't need to see the <nowiki> tag.
     if ($tagname eq 'nowiki')
     {
-	push @$tagstack, $tagname;
+	push @$tagstack, $tagname
+	    unless $isEmptyTag;
 	return;
     }
 
@@ -735,7 +749,7 @@ sub _html_tag
 		    }
 	    }
     }
-    $newtag .= " /" if $HTML::Tagset::emptyElement{$tagname};
+    $newtag .= " /" if $HTML::Tagset::emptyElement{$tagname} || $isEmptyTag;
     $newtag .= ">";
 
     # If this isn't a block level element, there's no need to track nesting.
@@ -749,7 +763,7 @@ sub _html_tag
     # Some elements can close implicitly
     if (@$tagstack)
     {
-	if ($tagname eq $tagstack->[-1]
+	if ($tagname eq $stacktop
 	    && $HTML::Tagset::optionalEndTag{$tagname})
 	{
 	    pop @$tagstack;
@@ -763,7 +777,7 @@ sub _html_tag
 		last if grep /^\Q$checking\E$/,
 			@HTML::Tagset::p_closure_barriers;
 
-		if ($tagstack->[$i] eq 'p')
+		if ($checking eq 'p')
 		{
 		    # pop 'em all.
 		    splice @$tagstack, $i;
@@ -777,9 +791,11 @@ sub _html_tag
     # they belong.
 
     # Push the new tag onto the stack.
-    push @$tagstack, $tagname;
+    push @$tagstack, $tagname
+	unless $isEmptyTag;
 
-    _append_processed_line $parser, $newtag, 'html';
+    _append_processed_line $parser, $newtag,
+			   $tagname eq 'pre' ? 'nowiki' : 'html';
     return;
 }
 
@@ -796,18 +812,14 @@ sub _html_text
     my $tagstack = $parser->{tag_stack};
     my ($newtext, $newstate);
 
-    print STDERR "Got skipped_text: `$skipped_text'\n"
+    warnings::warnif ("Got skipped_text: `$skipped_text'")
 	if $skipped_text;
 
     if (@$tagstack)
     {
-	if ($tagstack->[-1] eq 'nowiki')
+	if (grep /\Q$tagstack->[-1]\E/, qw{nowiki pre})
 	{
-	    $newstate = 'nowiki';
-	}
-	elsif ($tagstack->[-1] eq 'pre')
-	{
-	    $newstate = 'html';
+	    $newstate = 'nowiki'
 	}
 	elsif ($is_cdata && $HTML::Tagset::isCDATA_Parent{$tagstack->[-1]})
 	{
@@ -820,7 +832,9 @@ sub _html_text
     unless (defined $newtext)
     {
 	$newtext = CGI::escapeHTML $dtext unless defined $newtext;
-	# CGI::escapeHTML escapes single quotes for some reason.
+	# CGI::escapeHTML escapes single quotes so the text may be included
+	# in attribute values, but we know we aren't processing an attribute
+	# value here.
 	$newtext =~ s/&#39;/'/g;
     }
 
@@ -1068,45 +1082,61 @@ sub _find_links
 {
     my ($text, $tags, $opts) = @_;
 
-    my ($s, $f);
+    # Build Regexp
+    my @res;
+
     if ($opts->{absolute_links})
     {
+	# URI
+	my $s;
 	$tags->{_schema_regex} ||= _make_schema_regex @{$tags->{schemas}};
 	$s = $tags->{_schema_regex};
+	push @res, qr/\b$s:[$uricCheat][$uric]*/
     }
 
-    if (ref $tags->{extended_link_delimiters} eq "ARRAY")
+    if ($opts->{implicit_links})
     {
-	# Backwards compatibility for absolute links
-	$text =~ s/(?:&lt;)?\b$s:[$uricCheat][$uric]*(?:&gt;)?/
-		   $tags->{link}->($&, $opts, $tags)/egi
-	    if $opts->{absolute_links};
-
-	my ($start, $end) = @{$tags->{extended_link_delimiters}};
-	while (my @pieces = _find_innermost_balanced_pair ($text, $start, $end))
+	# StudlyCaps
+	if ($tags->{implicit_link_delimiters})
 	{
-	    my ($tag, $before, $after) = map { defined $_ ? $_ : '' } @pieces;
-	    my $extended               = $tags->{link}->( $tag, $opts, $tags )
-					 || '';
-	    $text                      = $before . $extended . $after;
+	    push @res, qr/$tags->{implicit_link_delimiters}/;
+	}
+	else
+	{
+	    warnings::warnif ("Ignoring implicit_links option since implicit_link_delimiters is empty");
 	}
-
-	# Backwards compatibility for implicit links
-	$text =~ s|(?<!["/>=])\b(?:[A-Z][a-z0-9]\w*){2,}|
-		   $tags->{link}->($&, $opts, $tags)|eg
-	    if $opts->{implicit_links};
     }
-    else
-    {
-	# Build Regexp
-	my @res;
-	push @res, qr/(?:&lt;)?\b$s:[$uricCheat][$uric]*(?:&gt;)?/
-	    if $opts->{absolute_links};		# URL
-	push @res, qr/$tags->{extended_link_delimiters}/
-	    if $opts->{extended};		# [[Wiki Page]]
-	push @res, qr/$tags->{implicit_link_delimiters}/
-	    if $opts->{implicit_links};		# StudlyCaps
 
+    if ($opts->{extended})
+    {
+	# [[Wiki Page]]
+	if (!$tags->{extended_link_delimiters})
+	{
+	    warnings::warnif ("Ignoring extended option since extended_link_delimiters is empty");
+	}
+	elsif (ref $tags->{extended_link_delimiters} eq "ARRAY")
+	{
+	    # Backwards compatibility for extended links.
+	    # Bypasses the regex substitution used by absolute and implicit
+	    # links.
+	    my ($start, $end) = @{$tags->{extended_link_delimiters}};
+	    while (my @pieces = _find_innermost_balanced_pair ($text, $start,
+							       $end))
+	    {
+		my ($tag, $before, $after) = map { defined $_ ? $_ : '' }
+						 @pieces;
+		my $extended = $tags->{link}->($tag, $opts, $tags) || '';
+		$text = $before . $extended . $after;
+	    }
+	}
+	else
+	{
+	    push @res, qr/$tags->{extended_link_delimiters}/;
+	}
+    }
+
+    if (@res)
+    {
 	my $re = join "|", @res;
 	$text =~ s/$re/$tags->{link}->($&, $opts, $tags)/ge;
     }
@@ -1167,7 +1197,8 @@ override the defaults.  This code:
 	use Text::MediawikiFormat as => 'wf', extended => 0;
 	wf ('some text', {}, {extended => 1});
 
-enables extended links, though the default is to disable them.
+enables extended links, after specifying that the default behavior should be
+to disable them.
 
 =head1 GORY DETAILS
 
@@ -1396,7 +1427,7 @@ I<Sic transit gloria mundi>.
 
 =head1 COPYRIGHT & LICENSE
 
- Copyright (c) 2006 Derek R. Price, all rights reserved.
+ Copyright (c) 2006-2008 Derek R. Price, all rights reserved.
  Copyright (c) 2002 - 2006, chromatic, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
